% File supr3/man/cluster.Rd
\name{3. Cluster}
\alias{collect}
\alias{cluster}
\alias{cluster.eval}
\alias{cluster.get}
%\alias{cluster.put}
%\alias{job.get}
%\alias{job.isDone}
%\alias{job.cancel}
%\alias{job.info}
%\alias{job.rm}
\alias{cluster.combine}
\alias{combine}
\alias{sync}
\alias{wait}
\alias{nextSubset}
\alias{.job.id}
\alias{.tr.id}
\alias{notify}

\title{SupR Cluster Interface}
\description{
  Functions to interface with the SupR cluster system.
%  These include those that can obtained by \code{grep("job|[Cc]luster", ls("package:supr2"), value=TRUE)}.
}
%cluster.put(uri, name, value)
\usage{

.GlobalEnv$.SuprEnv$.LastValue
.job.id
.tr.id
JobEnv

clusterenv()
clusterenv()$jobs()
clusterenv()$job(job_id)

cluster.eval(expr, data, envir = list(), error.handler = NULL,
	     env = parent.frame(), wait = TRUE, combine = job.combine.default,
	     ...) 
cluster.combine(x, combine, bykey = FALSE, env = parent.frame())
cluster.get(uri, name, remove = FALSE)
collect(x, remove = FALSE)

nextSubset(env = parent.frame())
info(text, conn = NULL) 
# CHANGE IT TO cluster.combine

sync(monitor, expr, language.only = FALSE, env = parent.frame())
wait(monitor, timeout = 0, env = parent.frame())
notify(monitor, all = FALSE, env = parent.frame())

broadcast(..., addr = NULL)
broadcasted(new = FALSE, timeout = 0)
}
\arguments{
  \item{expr}{a meaningful expression executable by \code{\link{Taskrunners}}.}
  \item{data}{either a list object of subsets or a distributed data in 
	  \code{SDFS}.}
  \item{envir}{an environment objection used by \code{\link{Taskrunners}} to
	  to evaluate \code{expr} for subsets. See \code{\link{Details}}.} 
  \item{env}{an environment object.}
  \item{error.handler}{ the error handler at the cluster level (not implemented yet)}
  \item{wait}{a logical object. }
  \item{combine}{a function used \code{cluster.eval} to combine results from taskrunners or a function that is to be used in \code{cluster.combine} to combine.}
  \item{uri}{a URI object. See \code{\link{Details}}.}
  \item{name}{a character string used as a name.}
  \item{job_id}{an integer used as job id.}
  \item{timeout}{an amount of time to wait for the job to finish.}
  \item{remove}{a logical object.}
  \item{bykey}{a logical object.}
  \item{x}{a named list or environment object.}
  \item{...}{optional arguments.} For \code{cluster.eval}, they are job-specific
  variables defined at the cluster level, i.e., at the \code{driver} site.
  For \code{broadcast}, they are variables to be sent to taskrunners.
}
\details{

% The \code{worker.options()} function allows the user to set and examine worker options. The available option names include code{ntrs} (for the number of taskrunners).  Numerical values of \code{ntrs} in the range from 0 to 1 are interpreted as fractions of the total number of computer cores at the worker node.
 
% The \code{dcl}, \code{dcl.expr},  \code{dcl.sync}, and \code{dcl.envir} arguments are used to specify a data-change listener. The listener is a process-based thread and is identified by the character value of \code{dcl}. While taskrunners are all evaluating \code{expr}, the data-change listener is evaluating \code{dcl.expr} with \code{dcl.envir} on its search path. The \code{dcl.sync} option is set TRUE, the evaluation of the data-change listener is synchronized with the evaluations of the taskrunners.


 The object \code{JobEnv} acts as a \code{UserDatabase} linked to a cluster
 environment object at the \code{driver} session. 
 The variables \code{job}, \code{task}, \code{address}, 
 \code{firstSubset}, and \code{.__NAMESPACE__.} are handled locally.
 The variable job is an integer object of a job id.
 When the value of \code{job} is \code{-1}, \code{JobEnv} is referred to as
 a \code{UserDatabase} to access global environment for job variables
 located at the \code{driver} site.

...
}
\value{
\code{cluster.eval} returns the evaluated value for \code{wait = TRUE},
otherwise, it returns an object of the \code{Future} class.
The returned result is a list of TaskrunnerValue objects from individual
\code{\link{taskrunners}}. A tr.val (Taskrunner Value) object is a list
consisting of
a \emph{value} component, a \emph{warnings} component,
and \emph{taskrunner} name. IN case of any error, the \emph{value}
is replaced by an \emph{error} component.
See \link{Examples}.

A job state notification is sent to \code{user}
as a message (see \code{send}, \code{recv}, and \code{messages})
by the \code{driver}. The message is an object of the \code{JobStatus} class.
Objects of the \code{JobStatus} class have
three named objects, namely, \code{job}, \code{state}, and \code{ntrs}.
The \code{job} element is an \code{integer} as the job id.
The \code{ntrs} element is the number of task runners that are still involved in
finishing the job.
The \code{state} element takes one of the three values as follows:

  \tabular{rlll}{
      \tab Job State \tab Notification Value \tab Comments\cr
    . \tab -------------------------- \tab --------------------- \tab --------------------------------------\cr
    . \tab TASKS_ASSIGNED  \tab \code{"assigned"}   \tab all the subsets have been taken\cr
    . \tab CANCELLED  \tab \code{"cancelled"}   \tab by user or due to errors\cr
    . \tab FINISHED  \tab \code{"finished"}   \tab \cr
    . \tab -------------------------- \tab --------------------- \tab --------------------------------------\cr
  }

With \code{wait = TRUE}, \code{cluster.eval} returns an
environment object, having a S3 class name \code{"Future"}.
The class \code{"Future"} has the following objects to
operate on the submitted job.

\preformatted{
class "Future" \{
    cancel(): cancel the submitted job.
    get(timeout = 0): get the result if available within next \code{timeout}
                      seconds.
    is.cancelled(): returns TRUE if the job is cancelled and FALSE otherwise.
    is.done(): return TRUE if the job is finished and FALSE otherwise.
    job.id: an integer object used as the job id.
    .value: the list of returned values from the taskrunners.
    combine: combine the list of taskrunner results, \code{.value}.
    .done: a logical object used internally by the \code{is.done} function.
    envir: the environment object of the \code{env} argument in calling
           \code{cluster()$eval}.
\}
}



When a job submitted with \code{wait = TRUE} is user-interrupted by, for example,
pressing Ctr-C on the keyboard, the job will continue as with \code{wait = FALSE}.
The object of class \code{"Future"} for accessing this job can be obtained
from the current value of  \code{.GlobalEnv$.SuprEnv$.LastValue}.





%\code{nextEvent} returns a named object.  When the value of a variable in the job environment, JobEnv, is changed (or set), an event is generated. The corresponding value returned by \code{nextEvent} has the name and value of the variable as its name and value.  When the count of a condition variable becomes zero (see \link{cond.count}), a list-valued event, (named \code{"condition"}), is generated.  The elements of the list are those that are provided by taskrunners via the \code{cond.count()} function.  When the job is complete, a NULL-valued event without a name is generated.  See \link{Examples}.

The three objects named \code{.job.id}, \code{.tr.id}, \code{.tr.name}
are defined in the environment, \code{.GlobalEnv$.SuprEnv}, of taskrunners.
The value of \code{.job.id} object is the integer value of the 
ID of the job the taskrunner is running.
The value of \code{.tr.id} object is the integer value of the 
job-specific taskrunner ID assigned to the taskrunner running the job.
The value of \code{.tr.name} object is the character of length one 
to uniquely identify the taskrunner process. Its value has the form of
hostname:pid and equals
\code{paste(Sys.info()["nodename"], Sys.getpid(), sep=":")}.

%The value of the symbol \code{.LastJob} is the last submitted job.

}
\note{
	...
}
\seealso{
  \code{dfs},
  \code{thread}
}
\examples{

\dontrun{

  library(supr3)
  cluster <- clusterenv()
  supr.options(level = 1)

  ls(cluster)
  ls(cluster$.LocalEnv)
  cluster$connections
  cluster$pthreads
  cluster$supr.options()
  cluster$supr.options(level=1)
}

# An echo example of different kinds
\dontrun{

## a plain version
  expr.echo <- quote({
     subsets <- list()
     while(!is.null(x <- nextSubset()))
        subsets <- append(subsets, x)
     subsets
  })
  results <- cluster.eval(expr.echo, as.list(1:100), envir=list())
  sort(unlist(results))
  messages()


## a version making use of the send function 
  expr.echo <- quote({
     subsets <- list()
     while(!is.null(x <- nextSubset())){
        send(x, "user")
        subsets <- append(subsets, x)
     }
     subsets
  })
  results <- cluster.eval(expr.echo, as.list(1:100))
  sort(unlist(results))
  messages()


## a version making use of the send and recv functions
  expr.echo <- quote({
     subsets <- list()
     while(!is.null(x <- nextSubset())){
        Sys.sleep(1)
        send(x, "user")
        subsets <- append(subsets, x)
     }
     subsets
  })

  eval({
    results <- cluster.eval(expr.echo, as.list(1:100), wait=FALSE)

    i <- 0
    y <- numeric(0)
    while(!is.null(x <- recv(timeout=60))){
      if(class(x) == "JobStatus"){
        print(x)
        if(x$state == "finished") break;
      } else {
        cat("[", i <- i + 1,"] x:", x, "\n")
        y[i] <- x
        plot(1:i, y)
      }
    }

    sort(unlist(results$get()))
  })

}


## a version making use of JobEnv
\dontrun{
  library(supr3)
  cluster <- clusterenv()
  supr.options(level=1)

  expr.echo <- quote({
     subsets <- list()
     while(!is.null(x <- nextSubset())){
        JobEnv$z <- x
        Sys.sleep(1)
        info(paste("JobEnv$i =",JobEnv$i));
        send(x, "user")
        subsets <- append(subsets, x)
     }
     subsets
  })

  eval({
    results <- cluster.eval(expr.echo, as.list(1:100), wait=FALSE)

    JobEnv$job  <- results$job.id
    JobEnv$task <- -1
    i <- 0
    y <- numeric(0)
    z <- numeric(0)
    while(!is.null(x <- recv(timeout=60))){
      if(class(x) == "JobStatus"){
        print(x)
        if(x$state == "finished") break;
      } else {
        cat("[", i <- i + 1,"] x:", x, "\n")
        JobEnv$i <- i
        y[i] <- x
        z[i] <- JobEnv$z
        plot(1:i, y)
        lines(1:i, z)
      }
    }

    sort(unlist(results$get()))
  })

  cluster$jobs()
  context(shutdown)
}


\dontrun{
# The use of the parallel random number generator in the rlecuyer package
  library(supr3)
  cluster <- clusterenv()
  supr.options(level=1)

  expr.rgn <- quote({ # FIXME?
          library(rlecuyer)
	  retval <- list()
          while(!is.null(x <- nextSubset())) {
             info(paste("x:", x, "pid:", Sys.getpid()))
             .lec.SetPackageSeed()
             .lec.StreamNames <- paste("lec.stream", 1:ntasks, sep=".")
             .lec.CreateStream(.lec.StreamNames)
             kind.old <- .lec.CurrentStream(.lec.StreamNames[.tr.id])

             val <- list(ntasks=ntasks, x=x,tr.id=.tr.id,
			    u.1 = runif(1), u.2 = runif(1))

             .lec.CurrentStreamEnd(kind.old)
             .lec.DeleteStream(.lec.StreamNames)
             retval <- rbind(retval, val)
	  }
          info(paste("x:", x, "pid:", Sys.getpid()))
	  retval
       })

  results <- cluster.eval(expr.rgn, as.list(1:100L),
                         envir=list2env(list(ntasks=100L)), wait=F)

}

# A cluster-level combine example
\dontrun{

  library(supr3)
  cluster <- clusterenv()
  supr.options(level=1)

  expr.combine <- quote({
    my.combine <- function(x, y){x+y}
    x <- nextSubset()
    while(!is.null( y <- nextSubset() )){
           x <- my.combine(x, y)
           Sys.sleep(1)
    }
    cluster.combine(x, my.combine)
  })

  results <- cluster.eval(expr.combine, data=as.list(1:100))

  results
  unlist(results)

  cluster$jobs()
  context(shutdown)
}

# A cluster-level combine_by_key example

\dontrun{

  library(supr3)
  cluster <- clusterenv()
  supr.options(level=1)

  data <- 1:100
  data <- lapply(as.list(data), function(x) {
                     names(x) <- LETTERS[(x-1) \%\% 26 + 1]
                     x
                  })

  expr.cbk <- quote({
    Sys.sleep(1)
    tr.combine <- function(x, y){
       keys <- unique(c(names(x), names(y)))
       z <- numeric(length(keys))
       names(z) <- keys
       for(key in keys)
	 z[key] <- sum(c(if(is.null(x)) x else x[names(x)==key],
			 if(is.null(y)) y else y[names(y)==key]))
       z
    }
    y <- NULL
    while(!is.null(x <- nextSubset()))
        y <- tr.combine(y, x)

    data <- as.environment(as.list(y))
    send(ls.str(data), "user")
    my.combine <- function(x, y) x+y
    cluster.combine(data, my.combine, bykey = TRUE)
  })

  results <- cluster.eval(expr.cbk, data=data)

  messages()

  r <- collect(results, TRUE)
  r <- unlist(r)
  
  # Checking
  s <- sapply(LETTERS, function(x) {
              sum(unlist(lapply(data, function(y) y[names(y)==x])))
       })
  s <- s[sort(names(s))]

  # Checking: is r[names(s)] - s a vector of 26 zeros?
  print(r[names(s)] - s)


}



# A simple EM example using the standard synchronization method.
# For the DD object EM_data, see the DD example
\dontrun{
  options(verbose=TRUE)
  options(debug=TRUE)

#  options(info=TRUE) # info default is TRUE
  library(supr3) # .Call("Pthread_sigmask", NULL)
  context(driver = context()$driver)
  cluster <- context()$driver
  cluster$connections

  expr.em <- quote({ subsets <- list()
    while(!is.null(sub <- nextSubset())) { x <- sub$x[,1]
      subsets <- append(subsets, list(x)); Sys.sleep(0.1)
# if(runif(1)<0.01) stop("testing")
    }
    n.subsets = length(subsets)
    if(n.subsets==0){ stop("FIXME"); X <- numeric(0); mu <- 0 } else {
      X <- unlist(subsets); mu <- mean(X[!is.na(X)])
    } 
    print(length(X)); print(head(X)); print(typeof(X)); print(class(X))

    mis <- is.na(X); n.com <- length(X)
    E.step <- function(mu){
      X[mis] <- mu
      list(sx = sum(X), n = n.com)
    }
    my.combine <- function(x, y){ list(sx = x$sx + y$sx, n = x$n + y$n) }
    M.step <- function(SS){ SS$sx/SS$n }

    n <- 0; max.n = 10; this.job <- paste0("JOB-", .job_id)
    while(n < max.n) { n <- n  + 1
      SS <- E.step(mu) #my.result <- cluster.combine(SS, my.combine)
      my.result <- cluster.combine(SS, my.combine)
      if (is.null(my.result)) {
        mu <- sync(std_EM, {timeout <- wait(std_EM, timeout=timeout())
		   if(timeout < 0) notify(std_EM, all=TRUE)
		   JobEnv$mu })
 #if(runif(1)<0.01) stop("testing") else if(runif(1)<0.01) q("no")
      } else {
        mu <- M.step(my.result) #broadcast(mu=mu, test="OKAY?")
	JobEnv$mu <- mu 
        sync(std_EM, { timeout <- wait(std_EM, timeout=timeout())
		    if(timeout < 0) notify(std_EM, all=TRUE)
		   s <- if(timeout < 0) "*" else " "
        #send(paste(s, "job.id =", .job_id, "iter =", n, "mu =", mu), "user")
        info(paste(s, "job.id =", .job_id, "tr.id =", .tr_id,  "iter =", n, "mu =", mu))
		   })
      }
    }
    list(mu=mu, Sx=SS$sx, n = n, n.mis=sum(mis), n.subsets=n.subsets)
  })
  print(proc.time())
  result <- cluster.eval(expr.em, data=structure("EM_data", class="DD"))
  print(proc.time())

  for(i in 1:5000){print(i); result <- cluster.eval(expr.em, data=structure("EM_data", class="DD"), wait=TRUE) }


  for(i in 1:10){print(i); result <- cluster.eval(expr.em, data=structure("EM_data", class="DD"), wait=FALSE) }

  # CHECKING job_id > 60?
  options(info=TRUE)
  library(supr3)
  context(driver = context()$driver)
  cluster <- context()$driver
  cluster$connections


  ## Use broadcast for sync
  expr.em <- quote({ subsets <- list()
    while(!is.null(sub <- nextSubset())) { x <- sub$x[,1]
      subsets <- append(subsets, list(x)); Sys.sleep(0.1)
    }
    n.subsets = length(subsets)
    if(n.subsets==0){ stop("FIXME"); X <- numeric(0); mu <- 0 } else {
      X <- unlist(subsets); mu <- mean(X[!is.na(X)])
    } 
    print(length(X)); print(head(X)); print(typeof(X)); print(class(X))

    mis <- is.na(X); n.com <- length(X)
    E.step <- function(mu){
      X[mis] <- mu
      list(sx = sum(X), n = n.com)
    }
    my.combine <- function(x, y){ list(sx = x$sx + y$sx, n = x$n + y$n) }
    M.step <- function(SS){ SS$sx/SS$n }

    n <- 0; max.n = 50; this.job <- paste0("JOB-", .job_id)
    while(n < max.n) { n <- n  + 1
      SS <- E.step(mu) #my.result <- cluster.combine(SS, my.combine)
      my.result <- cluster.combine(SS, my.combine)
      if (!is.null(my.result)) {
        mu <- M.step(my.result)
	broadcast(mu=mu, iter = n)
        info(paste("pid =", Sys.getpid(), "job.id =", .job_id,
		   "tr.id =", .tr_id, "iter =", n, "mu =", mu))
#	info(as.character(rc))
      }
#      if(runif(1)<0.01) stop("testing")
      mu <- broadcasted(new = TRUE)$mu
    }
    list(mu=mu, Sx=SS$sx, n = n, n.mis=sum(mis), n.subsets=n.subsets)
  })

  print(proc.time())
  result <- cluster.eval(expr.em, data=structure("EM_data", class="DD"))
  print(proc.time())


  i <- 0; while(TRUE){i <- i+1; print(i); result <- cluster.eval(expr.em, data=structure("EM_data", class="DD")) }


  library(supr3)
  cluster <- clusterenv()
  supr.options(level=1)

  ## Use broadcast for sync
  expr.em <- quote({ subsets <- list()
    while(!is.null(sub <- nextSubset())) { x <- sub$x[,1]
      subsets <- append(subsets, list(x)); Sys.sleep(0.1)
    }
    n.subsets = length(subsets)
    if(n.subsets==0){ stop("FIXME"); X <- numeric(0); mu <- 0 } else {
      X <- unlist(subsets); mu <- mean(X[!is.na(X)])
    } 
    print(length(X)); print(head(X)); print(typeof(X)); print(class(X))

    mis <- is.na(X); n.com <- length(X)
    E.step <- function(mu){
      X[mis] <- mu
      list(sx = sum(X), n = n.com)
    }
    my.combine <- function(x, y){ list(sx = x$sx + y$sx, n = x$n + y$n) }
    M.step <- function(SS){ SS$sx/SS$n }

    n <- 0; this.job <- paste0("JOB-", .job_id)
    while(n < max.n) { n <- n  + 1
      SS <- E.step(mu) #my.result <- cluster.combine(SS, my.combine)
      my.result <- cluster.combine(SS, my.combine)
      if (!is.null(my.result)) {
        mu <- M.step(my.result)
	broadcast(mu=mu, iter = n)
        send(list(mu=mu, n=n), "user")
      }
      mu <- broadcasted(new = TRUE)$mu
    }
    list(mu=mu, Sx=SS$sx, n = n, n.mis=sum(mis), n.subsets=n.subsets)
  })

  messages()
  res <- eval({
    result <- cluster.eval(expr.em, data=structure("EM_data", class="DD"),
		   wait = FALSE, envir=list(max.n = 50), my.par=1:10)
    i <- 0; mu <- n <- numeric(0)
    while(!is.null(x <- recv(timeout=5))){
      if(class(x) == "JobStatus"){
        print(x)
        if(x$state == "finished") break;
      } else {
        i <- i+1
        mu[i] <- x$mu
        n[i] <- x$n
        s <- ceiling(length(n)/2):length(n)
        plot(n[s], mu[s])
      }
    }
    result$get()
  })

  messages()


}

## An asynchronous distributed EM
\dontrun{
  library(supr3)
  cluster <- clusterenv()
  supr.options(level=1)

  expr.em <- quote({ subsets <- list()
    while(!is.null(sub <- nextSubset())) { x <- sub$x[,1]
      subsets <- append(subsets, list(x)); Sys.sleep(0.1)
    }
    n.subsets = length(subsets)
    if(n.subsets==0){ stop("FIXME"); X <- numeric(0); mu <- 0 } else {
      X <- unlist(subsets); mu <- mean(X[!is.na(X)])
    } 
    print(length(X)); print(head(X)); print(typeof(X)); print(class(X))

    mis <- is.na(X); n.com <- length(X)
    E.step <- function(mu){
      X[mis] <- mu
      list(sx = sum(X), n = n.com)
    }
    my.combine <- function(x, y){ list(sx = x$sx + y$sx, n = x$n + y$n) }

    n <- 0; this.job <- paste0("JOB-", .job_id)
    while(n < max.n) { n <- n  + 1
      SS <- E.step(mu)
      send(list(SS=SS, tr = .tr_id), "user")
      mu <- broadcasted(new = TRUE)$mu
      info(paste("mu:", mu))
    }
    list(mu=mu, Sx=SS$sx, n = n, n.mis=sum(mis), n.subsets=n.subsets)
  })

  messages()
  res <- eval({

    M.step <- function(SS){ 
	    sum(unlist(lapply(SS, "[[", "sx")))/
	    sum(unlist(lapply(SS, "[[", "n")))
    }
    result <- cluster.eval(expr.em, data=structure("EM_data", class="DD"),
		   wait = FALSE, envir=list(max.n = 50), my.par=1:10)

    JobEnv$job  <- result$job.id; JobEnv$task <- -1
    ##ntrs <- 0; SS <- list(); R <- logical(0)

    n <- 0; mu.seq <- n.seq <- numeric(0)
    while(!is.null(x <- recv(timeout=5))){
      if(class(x) == "JobStatus"){
	print(x)
       	if(x$state == "finished") break;
        if(x$state == "assigned") {
	      	ntrs <- x$ntrs
		SS <- lapply(as.list(1:ntrs), function(x) list(sx=0, n=0))
		R <- logical(ntrs)
		count <- logical(ntrs)
	}
      } else {
        k <- x$tr+1
        SS[[k]] <- x$SS
	R[k] <- TRUE
	if(sum(R) >= 0.75*ntrs || runif(1) < 0.01){
	  cat("sum(R):", sum(R), "\n")
	  mu <- M.step(SS)
	  n <- n + 1
	  broadcast(mu=mu, iter = n)
	  R[] <- FALSE
          mu.seq[n] <- mu
          n.seq[n] <- n
          s <- ceiling(length(n.seq)/2):length(n.seq)
          plot(n.seq[s], mu.seq[s])
	}
      }
    }
    if(is.null(x)) stop("unexpected NULL msg ...")
    # result$get()
  })

  # result$cancel()

}

%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
%
}



\keyword{thread}


