% File supr3/man/dd.Rd
\name{4. DDS or DFS}
\alias{DD}
\alias{DDT}
\alias{dfs}
\alias{sdfs}
\alias{DD.close}
\alias{DD.exists}
\alias{DD.get}
\alias{DD.list}
\alias{DD.mv}
\alias{DD.open}
\alias{DD.options}
\alias{DD.persist}
\alias{DD.put}
\alias{DD.replicate}
\alias{DD.rm}
\alias{DDEnv}
\alias{mv}
\alias{replicate}
\alias{persist}
\alias{data.table}
%  "get.DD"       "put.DD"

\title{SupR Distributed Data System}
\description{
  These functions represent an interface for accessing the SupR distributed
  data system (DDS).
  They include
 % those that can be obtained by \code{grep("dd|DD", ls("package:supr3"), value=TRUE)} and \code{grep("DDEnv", ls("package:supr3"), value=TRUE)}, and
 the common set of functions, namely,
  \code{assign} or \code{`$<-`},
  \code{exists},
  \code{get} or \code{`$`},
  \code{ls},
  and
  \code{rm}, as an interface for
  \code{environments} and
  \code{userDefinedDatabase} objects.
 % Methods for acting on DDS objects to extract and replace parts.
}
\usage{

DD(name)
DD.close(name, save = TRUE)
DD.exists(name, ...)
DD.get(name, subset.names, ..., env = parent.frame())
DD.list(name = ".")
DD.mv(src, dest)
DD.open(name, create = TRUE)
DD.options(...)
DD.persist(name, level = c("file", "tmp", "shm", "mem", "null"))
DD.put(dd.name, subsets, subset.names)
DD.replicate(name)
DD.rm(name, recursive = FALSE)

DDEnv(name=".", addr = context()$dfs)
open.DDEnv(ddenv, name, save = TRUE)
close.DDEnv(ddenv, name, save = TRUE)
is.DDEnv(x)
print.DDEnv(x, ...)
replicate.DDEnv(x)
mv(src, dest)

`[.DDEnv`(x, i, j, combine, ..., env = list())

}
\arguments{
  \item{create}{a logical object.}
  \item{name, dd.name}{a DD name, a character string representing a Unix file path.}
  \item{subset.names}{a character vector of Unix simple file names.}
  \item{env}{an environment for evaluations of arguments.}
  \item{src, dest}{DD names.}
  \item{level}{one of the values given in its default.}
  \item{subsets}{a list of objects representing subsets.}
  \item{subset.names}{a character vector specifying the subset names.}
  \item{...}{optional arguments.}
}
\details{

  The \code{DD()} function makes an DD object, regardless the existence of
      the named DD.

  The \code{DD.exists()} function tests for the existence of the distributed data. 

  The \code{DD.get()} function gets the named data subsets of the DD
      \code{name} from DFS.

  The \code{DD.list()} function returns simple information on
      the DD \code{name} or a file directory at DFS.

  The \code{DD.mv()} function rename \code{src} to \code{dest}.

  The \code{DD.open()} function opens the data data named \code{name} if it exists.
  Otherwise, it creates and opens the data if \code{create = TRUE},
   and generates an error if \code{create = FALSE}.

  The \code{DD.options()} invokes the \code{options()} function
    at DFS*.

  The \code{DD.persist()} changes the storage type of the named
    DD.

  The \code{DD.put()} function adds data subsets, a list of objects,
      with subset names \code{subset.names} to the distributed data
      named \code{name}. 
    
  The \code{DD.replicate()} increases the (minimum) number of 
     subset replications by one. 
    
  The \code{DD.rm()} removes the named DD.
}
\value{
  The \code{DD()} function returns an DD object, regardless the existence of
      the named DD.
	
  The \code{DD.close()} function returns an integer as the return code*
  The \code{DD.exists()} function returns TRUE if the named DD exists, and FALSE
    otherwise.
  The \code{DD.get()} function returns a list of objects.
  The \code{DD.list()} function returns a data frame object of
            subset name, size, timestamp, locations, and number of replications.
  The \code{DD.mv()} function returns an integer as the return code*
  The \code{DD.open()} function returns a DD object.

  The \code{DD.options()} function returns the value of the \code{options()} function
   call at DFS*.

  The \code{DD.persist()} function returns an integer as the return code*

  The \code{DD.put()} function returns the number of subsets.

  The \code{DD.replicate()} function returns the number of replications of the updated
     DD object.

  The \code{DD.rm()} function returns an integer as the return code*

  The \code{DDEnv} function returns a DDEnv object as an alternative
  representation of a DD object, in term of an
  \code{userDefinedDatabase} and \code{environment} object.
  The other functions following \code{DDEnv}
  on above usage list of functions are intended to provide
  operations to act on DD objects in the same way as
  on \code{environment} objects, as 
  demonstrated below in the examples.

  The \code{`[.DDEnv`} function provides a simple way to operate on DDEnv objects as distributed \code{data.frame} or \code{data.table}. The general form of
  DDEnv syntax is \code{DDEnv[i, j, combine, ..., env = list()]}.
  The \code{i} argument is an expression. When evaluated remotely, its value is
  either a \code{logical} or \code{integer} object, which is to be use as 
  to select rows of \code{DDEnv}.
  The \code{j} argument is an expression to be evaluated remotely.
  The \code{logical}, \code{integer}, and \code{character} values are
  used to select columns according to column variable positions and names.
  When its value is a list, a new \code{data.frame} or \code{data.table} type of
  object is created with the list of objects, when its components
 are objects of \code{logical}, \code{integer}, \code{numeric}, and
 \code{character} and have the same length.
}
\note{
	...
}
\seealso{
  \code{\link{Cluster}},
  \code{\link{Supr}},
  \code{\link{Thread}}.
}
\examples{

# Create a distributed data, named "EM_data", consisting of
# simulated subsets, and upload it into DFS
\dontrun{
  options(info=TRUE)
  library(supr3)

  context(dfs=context()$dfs)

  {
    data.name = "EM_data"
    DD.open(data.name)
    DD.list(data.name)
    tmp <- DD.get(data.name, NULL)
    length(tmp)
    table(is.na(tmp[[1]]$x))

    DD.close(data.name)

    supr.options(verbose=TRUE)

    x <- ddenv("EM_data")
    x[][]
  }


  ## distribute subsets of R objects saved in files
  data.name = "EM_data"

  if(DD.exists(data.name)) DD.rm(data.name, TRUE)

  DD.open(data.name)

  N <- 100L  # number of subsets
  K <- 1000L # subset size
  mis.prob <- 0.5  # fraction of missing values

  if(!dir.exists(data.name))
    dir.create(data.name)

  files <- paste0(data.name, "/part_", 1:N, ".rdata")
  for(i in 1:length(files)){
    x <- rnorm(K)
    mis <- runif(K) < mis.prob
    x[mis] <- NA
    x <- data.frame(x=x)
    save(N, x, file=files[i])
  }

  DD.put(data.name, files, NULL)
  DD.list(data.name)

  #unlink(data.name,  recursive = TRUE, force = TRUE)
}

## distribute R objects via R serialization
\dontrun{
  options(info=TRUE)
  library(supr3)
  context(dfs=context()$dfs)

  dn <- dnenv(1)
  dn$supr.options(level=1)
  dn <- dnenv(2)
  dn$supr.options(level=1)

  dfs <- context()$dfs
  nn <- nnenv()
  nn$supr.options(level=1)



  data.name = "data_frame"

  if(DD.exists(data.name)) DD.rm(data.name, TRUE)
  DD.open(data.name)
  N <- 10L  # number of subsets
  K <- 1000L # subset size
  P <- 10L # number of variables
  var.names <- paste("X", 1:P, sep=".")
  for(i in 1:N){
    X <- matrix(runif(K*P), nrow=K, dimnames=list(NULL, var.names))
    X <- as.data.frame(X)
    DD.put(data.name, list(X), paste0("part_", i))
    # added suffix ".sro" internally
  }

   DD.list(data.name)
   DD.persist(data.name)

   DD.get(data.name, "part_2.sro")

   A <- dfs$open(data.name)
   A[1:2, 2:4]

}


## Create one more replicate of the data in DFS
\dontrun{ DD.replicate(data.name)
  DD.list(data.name)
}

# Persist the data to disk
\dontrun{ DD.persist(data.name, "file")
  DD.close(data.name)
}

# Read data subsets, all subsets in this example
\dontrun{ tmp <- DD.get(data.name, NULL)
  length(tmp)
  table(is.na(tmp[[1]]$x))
  DD.close(data.name)
}

# Stop the SupR context
\dontrun{ context(shutdown)
}

# Double check
\dontrun{ context()

  Supr(verbose = TRUE)
  data.name = "EM_data"
  DD.open(data.name)
  tmp <- DD.get(data.name, NULL)
  length(tmp)
  table(is.na(tmp[[1]]$x))

  DD.close(data.name)

  context(shutdown)
}

\dontrun{
  # library(supr3)
  # context(dfs=context()$dfs)

  options(debug=TRUE)

  options(info=TRUE)
  library(supr3)
  context(dfs = context()$dfs)
  dfs <- context()$dfs
  dfs <- ddenv()

  {
    a <- dfsenv()
    a$connections
    a$pthreads
  }


  {
 	em.data <-  dfs$open("EMData")
  }

  context(dfs = NULL)


   em.data <- open(DDEnv("EMData"))

   assign("x", 1:10, envir = em.data)
   assign("y", 1:100, envir = em.data)
   ls(em.data)
   get("x", envir=em.data)
   get("y", envir=em.data)
   get("z", envir=em.data)
   em.data$x
   em.data$y
   em.data$z
   em.data$z <- 1:20
   em.data$z

   exists("EMData", envir=em.data) # check subsets... or no?

   rm(EMData, envir = em.data) # check subsets... or no?

   # mv(em.data, as.character(where))
   persist(em.data, level)
   replicate(em.data)


if(FALSE){

  ##############################################################
  ## some limited experiments on `[.DDEnv`                    ##
  ## a tada.table framework; See the data.table R package     ##
  ##############################################################

  options(info=TRUE)
  library(supr3)
  context(dfs = context()$dfs)
  #supr.options(level=1)

  dn <- dnenv(1)
  dn$supr.options(level=1)
  dn <- dnenv(2)
  dn$supr.options(level=1)

  dfs <- context()$dfs
  nn <- nnenv()
  nn$supr.options(level=1)

#  em.data <-  dfs$open("EMData")
  em.data <-  dfs$open("EM_data")

  b <- (a <- em.data[, list(x=x, y=x^2, z = foo(x), key = x > 0), , foo= sin])

  (a <- em.data[, list(x=x, y=x^2, z = foo(x), key = x > 0), test,
     test= function(x) {
    	  v <- x$value
          x <- x$x
	  if (is.null(x)) {
		  return(v)
	  }
	  else if (class(x) == "DD.frame") { # dn
		  n <- length(x)
		  append(v, as.list(n))
	  } else { # nn
		  append(v, x)
	  }
    }, foo= sin][,,dim])


  (a <- em.data[, list(x=x, y=x^2, z = foo(x), key = x > 0), test, test= function(x) {
    	c(x$value, runif(10))
    }, foo= sin])

  (a <- em.data[, list(x=x, y=x^2, z = foo(x), key = x > 0), dim , foo= sin])

  (a <- em.data[, list(x=x, y=x^2, z = foo(x), key = x > 0), , foo= sin][, 2:3, dim, j2=4])

  (a <- em.data[, list(x=x, y=x^2, z = foo(x), key = x > 0), , foo= sin][, 2:3, , j2=4][,,dim()])

  (a <- em.data[, list(x=x, y=x^2, z = foo(x), key = x > 0), , foo= sin][, 2:3, , j2=4][, , ][,,dim()])

  (a <- em.data[, list(x=x, y=x^2, z = foo(x), key = x > 0), function(x) {
	   if(runif(1)<0.1) stop("testing")
    	c(x$value, search())
    }, foo= sin])

  (a <- em.data[, list(x=x, y=x^2, z = foo(x), key = x > 0), function(x){if(runif(1) < 0.01) stop("TESTING"); 100} , foo= sin])


  (a <- em.data[!is.na(x), list(x=x, y=x^2, z = foo(x), key = x > 0), dim, foo= sin][, 1:3, , j2=4][, 1:2, group=function(){} ][1,2,3]
  )

  (a <- em.data[, list(x=x, y=x^2, z = foo(x), key = x > 0), .n, foo= sin][, 3, , j2=4][, , group=function(){} ][1,2,3]
  )

    a
}

   # close(em.data)
   {
     dfs$close(attr(em.data, "name"))
     em.data$close()
   }

   context(shutdown)
}


































}
\keyword{thread}

