% File supr2/man/supr.Rd
\name{2. SupR}
\alias{supr}
\alias{context}
\alias{supr.options}
\alias{ping}
\alias{masterenv}
\alias{workerenv}
\alias{nnenv}
\alias{dnenv}
\alias{send}
\alias{recv}
\alias{messages}


\title{Cluster, DFS, and Thread}
\description{
  SupR: an experimental R package for parallel and distributed computing.

  A cluster is referred to as a system consisting of a \code{master},
  a number of \code{worker}s, and a \code{driver}.
  The \code{master}  serves as a central server for workers and drivers.
  Workers are registered at the \code{master} and connect to
  \code{driver} once a driver is available.
  A \code{worker} manages a number of task-runners that are grouped under
  \code{executor}s. The current implementation of \code{worker} has a single
  \code{executor}. Each task-runner is a R session as a
  child process of the worker, and works with a \code{pthread} of
  the worker to interact with other components of the cluster.
 
  The \code{driver} provides a user interface to the cluster.
  Computation tasks are submitted to the cluster as jobs through the driver via
  the \code{cluster.eval} function.
  Details are  described in the \link{cluster} section.
}
%SuprContext(what, ...)
%SuprContext.conf(..., add = TRUE, save = FALSE, envir = parent.frame())
%ClusterEnv
%SharedEnv
\usage{
.SuprEnv
supr.version()
context(...)
supr.options(...)

suprenv(address)
masterenv()
driverenv()
workerenv(which)
nnenv()
dnenv(which)
%thenv()

ping(address, timeout = 10)
netstat(port, args = NULL)

send(msg, to)
recv(timeout = 0)
messages()

}
\arguments{
  \item{address}{a character object 
	  representing the socket server a SupR component; See \emph{Details}.}
  \item{which}{an integer object referring to the worker number registered at the \code{master} or the DFS datanode number at the DFS \code{namenode}; See \emph{Details}.}
  \item{...}{a list of named arguments reserved for future development.
  }
}
\details{

  The \code{.SuprEnv} variable is an environment object defined in
  \code{.GlobalEnv} in every \code{supr} component R session. It has \code{.GlobalEnv} as its parent environment,
  except \code{taskrunner}.
  It serves as the parent environment for cluster level evaluations.

  The \code{supr.version} function returns a character object of 
  the \code{supr} package version.

  The \code{context} function provides a simple way of working with SupR cluster context.
  It allow the user to start and examine components of a cluster.
  The arguments of \code{context} takes the following character (or symbol) values:

  \tabular{rlll}{
      \tab Argument \tab Value \tab Action\cr
    . \tab ----------- \tab ------------------------------ \tab -----------------------------------------------------\cr
    . \tab shutdown \tab   \tab command to shutdown cluster by\cr
    . \tab  \tab  \tab sending a shutdown message to \code{master},\cr
    . \tab  \tab  \tab \code{driver}, and \code{dfs_name}.\cr
    . \tab info  \tab  (integer) port_number \tab start an info socket server\cr
    . \tab dfs  \tab (character) socket_address \tab start DFS\cr
    . \tab  \tab  \code{NULL} \tab shutdown the dfs server\cr
    . \tab driver \tab (integer) port_number \tab start driver and cluster\cr
    . \tab  \tab  \code{NULL} \tab shutdown the driver\cr
    . \tab master \tab (character) socket_address \tab start master and workers\cr
    . \tab  \tab  \code{NULL} \tab shutdown the master\cr
    . \tab ths \tab (list) socket_addresses \tab start thread servers\cr
    . \tab  \tab  \tab return the context environment\cr
    . \tab  \tab  \tab session: luster component name\cr
    . \tab proc \tab \tab show all supr3 processes\cr
    . \tab killall \tab \tab kill all supr3 processes\cr
    . \tab config \tab a list of arguments \tab generate .supr/supr.conf\cr
    . \tab  \tab nodes= \tab a character object of hostnames\cr
    . \tab  \tab force= \tab TRUE: replace .supr/supr.conf  \cr
    . \tab  \tab        \tab FALSE: replace .supr/supr.conf when safe \cr
    . \tab  \tab port=   \tab lower limit on socket ports to use\cr
    . \tab  \tab ncpu=   \tab number of cpu corses/taskrunners\cr
    . \tab ----------- \tab ------------------------------ \tab -----------------------------------------------------\cr
  }

  The \code{supr.options} function has three \code{logical} options
  \code{info}, \code{verbose}, and \code{debug} and
  an \code{integer} option to control the system information details.
  It offers a \code{timeout} option that is used in establishing socket
  connections and reading from sockets.
  In addition, it has a \code{port} option, which
  is used as the lower limit of socket ports when a dynamic socket server port is allowed.
  

  The \code{suprenv} function creates a socket connection to the server of the cluster component. The target component is specified by its socket server address of the \code{hostname:port} format.
  The \code{driverenv}, \code{masterenv}, \code{workerenv}, \code{nnenv}, and
  \code{dnenv}
  functions are intended to provide a convenient way of accessing cluster components.

  The \code{ping} function sends an echo request to the socket server specified
  by the address argument.

  The \code{netstat} function uses the OS command \code{netstat} to provide
  information on network connections on a given port of interest.

  The \code{send} function sends an R object \code{msg} as a message to the recipient \code{to}. 

  The \code{recv} function reads one message, if there are messages available during the time period specified by the \code{timeout} (sec.) argument.

  The \code{messages} function reads all the messages that have been received from the last read.

%It returns an environment object as an interface to the component server.


%  The SupR configuration options are specified with one line for each of master, driver, worker, dfs_name, and dfs_data.  Each line consists of a name and a value separated by spaces and tabs.  The value is the server address of the component in the form of //hostname:port_number, such as \code{//localhost:0}.  When set 0, the port number \code{port_number}, to be searched dynamically in the range from 16384 to 49151.

%  For workers, the number of taskrunners can be specified via an optional setting in the form of \code{//hostname:port_number?ntrs=number_of_taskrunners}.

%  For dfs_data nodes, their default data directory is a subdirectory in the user directory.  Multiple data directories can be used.  In this case, they are specified by adding lines after the datanode specification, one line for each data directory using the syntax of the form:  \preformatted{ file://absolute_directory_path } More precisely, in terms of POSIX regular expressions, the syntax for each data directory can be expressed as \code{"[ \t]*file://[^\n]*"}.

  
  

%  \code{SuprContext.conf} returns the current context.
  

%  \code{ClusterEnv} is an active binding, returning an object of the \code{SuprShared} class. It is used to get and set objects, through the \code{`$`} and \code{`$<-`} operators, in the cluster environment. See \code{\link{examples}}.

%  \code{SharedEnv} is an active binding, returning an object of the \code{SuprShared} class. It is used to get and set objects through the \code{`$`} and \code{`$<-`} operators, for threads of three different types, namely, simple thread, local process-based, and (experimental) remote process-based, in a SupR (user) session.  See \code{\link{examples}}.

%  The access to objects in both \code{ClusterEnv} and \code{SuprShared} is automatically synchronized. The actual environment referenced by \code{SuprShared} is the object \code{SuprShared} in \code{.GlobalEnv}, whereas the actual environment referenced by \code{ClusterShared} is an environment object in the \code{\link{Driver}} session in an alive SuprContext.
}
\value{
  The \code{suprenv} function returns an environment object as an interface to the \code{.SuprEnv} object of the component server.
%  A few functions are also made available locally in the \code{.LocalEnv} environment variable that can be accessed from the returned environment object.
  Common objects that are available for all such environments include
  \code{.SuprThreads}, \code{connections}, \code{makeActiveBinding},
\code{pthreads}, \code{shm},    and \code{supr.options}.


  \code{masterenv()} returns an environment for accessing master's \code{.SuprEnv}.

  \code{driverenv()} returns an environment for accessing driver's \code{.SuprEnv}.
  In addition to the common objects listed above, this environment has 
  the following driver-specific objects:
  \code{dfs}, \code{master}, \code{job}, and \code{jobs}.


  \code{nnenv()} returns an environment for accessing DFS namenode's \code{.SuprEnv}.

  \code{workerenv(which)} returns an environment for accessing \code{.SuprEnv}
  of the which-th worker registered at the \code{master}.
  The call \code{workerenv(which=0)} returns all the worker addresses.

  \code{dnenv(which)} returns an environment for accessing \code{.SuprEnv}
  of the which-th datanode registered at the DFS \code{namenode}.
  The call \code{dnenv(which=0)} returns all the datenode addresses.

  \code{ping(address, timeout)} returns a received 
  integer object from the server before timeout is expired, and
  generates an error message otherwise.
  
  \code{netstat(port)} returns a \code{character} object as the information on socket connections involved the given \code{port}.

  \code{recv(timeout)} returns the next messages from the last read, if available before the timeout is expired, and \code{NULL} otherwise.

  \code{messages()} returns a list of messages.

}
\note{
The installation of the \code{supr} package should create the file named \code{.suprrc},
under user's home directory. This is to be run with the bash shell command
\code{source} to define three system environment variables
\code{R_HOME}, \code{SUPR_SYS_HOME}, and \code{SUPR_USR_HOME}.
The value of \code{R_HOME} is the \code{HOME} directory of the R version that
was used in the installation.
The value of \code{SUPR_SYS_HOME} is the \code{HOME} directory of the
installed \code{supr} package.
It is used for remote invocations of the supr commands 
\code{dfs_name},
\code{dfs_data},
\code{master},
\code{worker}, and
\code{threadserver}.
The value of \code{SUPR_USR_HOME}, with the default \code{$HOME/.supr}, is the
top-level parent directory
of the running \code{supr} commands.

Configurations of socket connection ports and distributed file locations
are specified in the configuration file: \code{$SUPR_USR_HOME/supr.conf}. Here is an example,
where a port number \code{n} followed immediately with '+' specifies any port in  the range from the port \code{n} is allowed.

\code{}

%\verb{}
\preformatted{
Namenode //machine.1:7200?data=\{
    file://$HOME/.supr/dfs_name/machine.1
\}

Datanode //machine.1:7201?data=\{
    file://$HOME/.supr/dfs_data/machine.1
    file://$HOME/SupR/dfs_data
\}

Datanode //machine.2:7201data=\{
    file://$HOME/.supr/dfs_data/machine.2
\}

Master //machine.1:7202

Driver //machine.1:7203

Worker //machine.1:7204
Worker //machine.2:7204

Infonode //machine.1:7205+

Filetransfer //machine.1:7206+

ThreadServers \{
    machine.1:7208+
    machine.2:7208+
\}

TrustedHosts \{
    machine.1
    machine.2
\}
}
Note that multiple files locations can be used for datanodes to store data.
This is illustrated with the first datanode in the above example.

The components of supr3 can also be started using machine operation system (OS)
commands:


\preformatted{
dfs_name [-port n] [--verbose] [--debug] [--info] [--help]
         [-info addr] [-notify addr]

dfs_data [-port n] [--verbose] [--debug] [--info] [--help]
         [-info addr] [-notify addr] [-nn addr]

master [-port n] [--verbose] [--debug] [--info] [--help]
       [-info addr] [-notify addr]

worker [-port n] [--verbose] [--debug] [--info] [--help]
       [-info addr] [-notify addr] [-master addr] [-ntrs k]

driver [-port n] [--verbose] [--debug] [--info] [--help]
       [-info addr] [-notify addr] [-shm name] [-master addr]
       [-dfs addr]


threadserver [-port n] [--verbose] [--debug] [--info] [--help]
       [-info addr] [-notify addr]

}

More information on these commands can be obtained by invoking them
with the \code{--help} argument.



%\deqn{
% \begin{array}{l}
%  \mbox{\tt Namenode //machine.1:7200?data= $\{$ }\\
%  \mbox{\tt \hspace{0.5in} file://$\$$HOME/.supr/dfs\_name/machine.1 }\\
%  \mbox{\tt $\}$}\\
% \end{array}
%}


}
\seealso{
  \code{\link{Cluster}},
  \code{\link{Thread}}
}
\examples{

## start
\dontrun{

##################################################
##### New functions, subject to more testing #####
##################################################

library(supr3)

context(proc)    # show SupR processes on all the nodes found in supr.conf
context(killall) # kill all SupR processes on all the nodes found in supr.conf

# The following function call

context(config=list(nodes="bell-fe[00-06]", force = TRUE, port=2022))

# generates ~/.supr/supr.conf with the given nodes argument and
# the same network domain as that of the host of the R session

# The next example shows how to use nodes obtained via
# an interactive Slurm job, sinteractive -N 10 -n 100 -A standby -t 1:0:0
# Slurm Reference: <https://slurm.schedmd.com/quickstart.html>


context(config= list(nodes = Sys.getenv("SLURM_JOB_NODELIST"),
         force=TRUE, port = 2022, ncpu=10))


# The following function call, subject to feature refinements, 

# workerenv(1)$supr.options(xterm = TRUE)
workerenv(1)$xterm(interactive=TRUE)

# starts a remote X Terminal (xterm) for stdout and stderr and
# an interactive pthread running R REPL (Read-Eval-Print Loop)
# on worker workerenv(1).

# Added Sys.info2

Sys.info2()

# which provides information on the number of cpu cores and the limits
# on the maximum file descriptor number that can be opened by
# current process

##################################################
#####          End of New functions          #####
##################################################



library(supr3)
supr.options()
supr.options(info = TRUE, verbose = TRUE)

# Show the initial state of the context
context()

# Start a cluster
context(driver = context()$driver)
cluster <- context()$driver
cluster$connections

masterenv()$connections
masterenv()$pthreads
workerenv(0)
workerenv(1)$connections
workerenv(1)$pthreads

nnenv()$connections
nnenv()$pthreads
dnenv(0) # datanode server addresses
dnenv(1)$connections
dnenv(1)$pthreads

# Shutdown the cluster
context(shutdown)
}

%
%
%
%
%
%
%
}
\keyword{thread}
\keyword{supr}


