% File supr3/man/thread.Rd
\name{5. Thread}
\alias{Thread}
\alias{ThreadServer}
\alias{thread}
\alias{ThreadServer.cancel}
\alias{ThreadServer.connect}
\alias{connect}
\alias{ThreadServer.start}
\alias{Thread.interrupt}
\alias{interrupt}
\alias{Thread.join}
\alias{join}
\alias{Thread.new}
\alias{Thread.start}
\alias{Thread.state}
\alias{state}
\alias{Sync.eval}
\alias{sync}
\alias{Sync.notify}
\alias{notify}
\alias{Sync.wait}
\alias{wait}

\title{SupR: Thread}
\description{
  SupR Multithreading with thread functions, implemented using both (Linux) processes and pthreads. To see the exported functions, run the code
  \code{grep("[Tt]hread", ls("package:supr3"), value=TRUE)}.
}
\usage{
ThreadServer.cancel(host, port, address = paste(host, port, sep = ":"))
ThreadServer.connect(host, port, address = paste(host, port, sep = ":"))
connect(host, port, address = paste(host, port, sep = ":"))
ThreadServer.start(port)

Thread.interrupt(thread)
interrupt(thread)
Thread.join(thread, useProxy = TRUE)
join(thread, useProxy = TRUE)
Thread.new(expr, ..., env = parent.frame(), error.handler = NULL, 
    server = local.thread.server, start = FALSE)
ThreadServer.connect(...)$new(expr, ..., env = parent.frame(), error.handler = NULL, 
    server = local.thread.server, start = FALSE)
start(thread)
state(thread)

Sync.eval(monitor, expr, language.only = FALSE, env = parent.frame())
Sync.notify(monitor, all = FALSE, env = parent.frame())
Sync.wait(monitor, timeout = 0, env = parent.frame())
sync(monitor, expr, language.only = FALSE, env = parent.frame())
notify(monitor, all = FALSE, env = parent.frame())
wait(monitor, timeout = 0, env = parent.frame())
}
\arguments{
  \item{env}{the parent of the environment where the \code{expr} is to be evaluated.}
  \item{error.handler}{a function to be called when errors occur. See
	 \code{\link{Details}}.}
  \item{expr}{an object to be evaluated by the thread. See 
	  \code{\link{Details}}.}
  \item{host}{a \code{\link{character}} representing the hostname of the thread server.}
  \item{port}{an \code{\link{integer}} representing the socket port of the thread server.}
  \item{monitor}{an expression of the form \code{\link{connection$name}} specifying a monitor named \code{\link{name}} at the ThreadServer \code{\link{connection}}.}
  \item{thread}{a \code{\link{Thread}} object created by \code{\link{Thread.new}}.}
  \item{start}{a \code{\link{logical}} object specifying that thread should
	  start to run right after it was created.}
  \item{...}{named objects. For \code{Thread.new}, they are passed on to the thread
	  environment.
} 
}
\details{
\code{ThreadServer.start} is used to start a thread server in the current R session.

\code{ThreadServer.connect} or \code{connect} is used to connect to the thread server specified by the arguments hostname \code{\link{host}} and socket port number \code{\link{port}}.

\code{ThreadServer.cancel}  cancels the thread server specified by the arguments hostname \code{\link{host}} and port number \code{\link{port}}.

\code{Thread.new} creates a new thread in the thread server session.
Non-name \code{\link{expr}} objects are to be evaluated by the thread.
Name \code{\link{expr}} objects are looked up in the current scope and
their bindings are to treated as non-name \code{\link{expr}} objects.


\code{Thread.start(thread)} starts the thread identified by the \code{thread} object.

\code{Thread.join(thread)} or \code{join(thread)} obtains the result of the \code{thread}. This should be always called because the system resources allocated for the thread can only be released via this function call, which invokes the C library function \code{waitpid} etc.

\code{Thread.interrupt(thread)} or \code{interrupt(thread)} terminates the running of the \code{thread}. 

\code{Thread.state(thread)} or \code{state(thread)} gets the last reported running state of the \code{thread}. 

\code{Sync.eval} obtains the monitor and evaluates the expression \code{expr} while holding the monitor.

\code{Sync.wait} or \code{wait} releases the monitor and waits to be notified or for a time period specified by the argument \code{timeout} in seconds. The special value \code{timeout = 0} means that no time limit is provided.
\code{timeout} can take an expression to be evaluated by the synchronization manager such as thread servers and driver in cluster evaluation. 
In this case, the value of the expression must be numeric and is used to determine the required time period (in seconds).

\code{Sync.notify} or \code{notify} wakes up threads waiting on the same monitor one at a time if  \code{all = FALSE} and all with a single call with \code{all = TRUE}.
}
\value{
\code{ThreadServer.start} returns a ThreadServer object, containing its hostname and port number for service connections.

\code{ThreadServer.connect} or \code{connect} returns a ThreadServer connection object. This object is to be used for the service of \code{\link{SupR-threading}}. It servers also as an \code{\link{environment}} object for accessing shared objects at the thread server session.

\code{ThreadServer.cancel} returns 0 if successful and -1 otherwise.

\code{Thread.new} return a thread identification number with the thread server attached as an attribute named \code{\link{server}}.

\code{Thread.start(thread)} returns an \code{\link{integer}} object 0 if successful and -1 otherwise.

\code{Thread.join(thread)} or \code{join(thread)} returns the value of thread evaluation along with the outputs to \code{stout} and \code{stderr} if the task was performed normally. When errors occur, the returned object has no value. In such a case, the error message, if available, is attached as an attribute named \code{error}.

\code{Thread.interrupt(thread)} or \code{interrupt(thread)} returns 0 if successful and -1 otherwise.

\code{Thread.state(thread)} or \code{state(thread)} returns an \code{\link{integer}} value as the state of the thread as follows: 0 --- created, 1 --- started, 2 --- evaluated, 3 --- finished.


\code{Sync.eval} returns the value of the expression \code{expr}.

\code{Sync.wait} or \code{wait} returns 0 if successful and -1 otherwise.

\code{Sync.notify} or \code{notify} returns 0 if successful and -1 otherwise.
}

\note{
The threading model here follows that of the JAVA language. Each thread is implemented with a Linux process.
 The process ID is used as the thread ID.

The shared objects at thread servers can be accessed via functions accessing objects in environments. A standard set of such functions consists of
\code{\link{ls}}, \code{\link{$}}, \code{\link{$<-}}, \code{\link{exists}}, and \code{\link{rm}}.
}

\references{
}


\seealso{
  \code{\link{SupR}} for some utility functions;
  \code{\link{Distributed File System}} for an alternative way of storing and accessing on SupR clusters; and
  \code{\link{Cluster}} for cluster such as map-and-reduce type of computing.
}
\examples{

### NEW ###
library(supr3)

# Generate supr.conf based on SLURM_JOB_NODELIST
# where the syntax of (the value of) SLURM_JOB_NODELIST is
# ...
# context(config= list(nodes = Sys.getenv("SLURM_JOB_NODELIST"),
#    force=TRUE, port = 2022), ncpu=5)

\dontrun{
supr.options(info=TRUE, xterm = TRUE)
context(killall)

# Start thread servers

context(ths=context()$ths)
ths <- context()$ths
ls(ths[[1]], all = TRUE)
ls(ths[[1]]$.LocalEnv, all = TRUE)
ths[[1]]$new

# Local info on thread servers
context(ths=list())

# Stop thread servers 
context(ths=NULL) # or context(shutdown)


}



### END_NEW ###

## Load the supr3 package in a R session:


library(supr3)

## Create a thread server in this the current session with a port number in an
## appropriate range.  Keep a record of this server address, consisting of the
## hostname and socket port number for connections from this and other R
## sessions on the same machine or different machines. At most one thread
## server can be running in a single R session. However, multiple thread
## servers can run simultaneously on different R sessions and on different
## machines.

\dontrun{
 server <- ThreadServer.start(port=2021L)
 print(server)

context(ths=context()$ths)
ths <- context()$ths

}

## Connect thread servers (only one server in the following example) from
## the same or different R session:

\dontrun{
servers <- attr(server, "address")

connections <- as.list(1:length(servers))
for(i in 1:length(servers)) 
    connections[[i]] <- connect(address = servers[i])


connections <- append(connections, ths)

}

## For a simple example, take one from these connections at random:

\dontrun{
conn <- connections[[sample(1:length(connections), size=1)]]
print(conn)

}

## Use the above connection conn as a shared network environment for R objects:

\dontrun{
ls(conn)
conn$x <- 1:10
ls(conn)
conn$x
exists("x", envir = conn)
exists("y", envir = conn)
rm("x", envir = conn)
objects(conn)

}

## Create a new thread to evaluate:
\dontrun{
s <- conn$new({
	 print(Sys.info()["nodename"])
	 print(Sys.getpid())
	 sin(1:10)
       })

start(s)
state(s)
res<- join(s)
print(res)

}

## or, equivalently,
\dontrun{
expr <- quote({
	 print(Sys.info()["nodename"])
	 print(Sys.getpid())
	 sin(1:10)
       })

s <- conn$new(expr)

start(s)
state(s)
res<- join(s)
print(res)

}


## Start the thread:
\dontrun{
start(s)

}

## Check the running state of the thread:
\dontrun{
state(s)

}

## Get the result and automatically release the resources:
\dontrun{
res <- join(s)

}


## A simple EM example to demonstrate the use of SupR thread synchronization.
## Create an artificial incomplete data from the standard Gaussian
## distribution with a fractional number of values missing at random:
\dontrun{
  n <- 1000; mis.frac <- 0.25
  X <- rnorm(n)
  X[runif(n) < mis.frac] <- NA

}

## Start default thread servers with addresses provided in .supr/supr.conf
\dontrun{
   library(supr3)
   supr.options(info=TRUE)
   #or
   supr.options(info=TRUE, xterm=TRUE)

   context(killall)

   context(ths=context()$ths)

   connections <- context()$ths

}


## Start m threads (at random) by the thread servers and choose a single
## server for shared objects
\dontrun{
m <- 10
shared <- connections[[sample(1:length(connections), size=1)]]
shared.addr <- attr(shared, "address")
# or
shared.addr <- shared$address

}

# Create an object named as 'SS' at the thread server 'shared' for storing
# the expected sufficient statistics, sum(X) and length(X):
\dontrun{
shared$SS <- c(0,0)

}

## Partition the data into m subsets:
\dontrun{
  m <- 10 # the number of threads to be used
  dim(X) <- c(m, n/m)

}

# Create m threads:
\dontrun{
  threads <- as.list(1:m)
  for(i in 1:length(threads)){
     conn <- connections[[sample(1:length(connections), size=1)]]
     expr <- quote({
	 shared <- connect(address = shared.addr)
         info("Start EM ...")
         which.mis <- which(is.na(x))
         for(it in 1:50){
             #E-step
             x[which.mis] <- mu
             my.SS <- c(sum(x), length(x))
             sync(shared$EM, {            #### a 'Sync.eval' example
                   SS <- shared$SS + my.SS
                   cat("[",it, "] SS:",SS, "n:", n, "\n")
                   if(SS[2] < n) {
                     shared$SS <- SS
                     wait(shared$EM)              #### a 'wait' example
                     mu <- shared$mu
                     cat("[",it, "] mu:",mu,"\n")
                   } else {
                     #M-step
                     mu <- SS[1]/n
                     info(paste("[",it, "] pid:",Sys.getpid(),
                        "mu:",mu))                #### send a text message
                     shared$mu <- mu
                     shared$SS <- c(0,0)
                     notify(shared$EM, all=TRUE)  #### a 'notify' example
                   }
                })
         }
         info(paste("\033[0;31m[",it, "] pid:",Sys.getpid(),"\033[0m"))
         shared$mu
     })
     threads[[i]] <- conn$new(expr, x=X[i,],n=n,mu=0,shared.addr=shared.addr)
  }

}

## Start the m threads
\dontrun{
for(i in 1:length(threads)){ start(threads[[i]]) }

for(i in 1:length(threads)){ cat("[", i,"] ", state(threads[[i]]), "\n") }

}

## Join the m threads
\dontrun{
results <- as.list(1:length(threads))
for(i in 1:length(threads)){ results[[i]] <- join(threads[[i]]) }

print(shared$mu)
print(mean(as.numeric(X), na.rm = TRUE))

}

## Cancel a thread server:
\dontrun{
 ThreadServer.cancel(address = attr(server, "address"))

}



}


\keyword{thread, concurrency}

